import math
import clingo
import random
import time


class modelResults:

    def addResults(self, model):
        self.cost = model.cost
        self.symbols = model.symbols(atoms=True)

def generateMaxiMinCycle(length, minval, maxval, agents):
    x = [f"#const n={agents}.", f"agent(1..n). resource(1..{length}). utils({minval}..{maxval}). ",
         "edge(A,B) :- resource(A), resource(B), B - A == 1.",
         "edge(B,A) :- B = #max{X:resource(X)}, A = #min{Y:resource(Y)}."]
    for i in range(1, agents + 1):
        for j in range(1, length + 1):
            v = random.randrange(minval, maxval)
            x.append(f"util({i},{j},{v}).")
    x = '\n'.join(x)
    return x


def generateMaxiMinComplete(size, minval, maxval, agents):
    x = [f"#const n={agents}.", f"agent(1..n). resource(1..{size}). utils({minval}..{maxval}). ",
         "edge(A,B) :- resource(A), resource(B), B != A, not edge(B,A)."]
    for i in range(1, agents + 1):
        for j in range(1, size + 1):
            v = random.randrange(minval, maxval)
            x.append(f"util({i},{j},{v}).")
    x = '\n'.join(x)
    return x

def generateMaxiMinGrid(minval, maxval, agents, x, y):
    encoding = [f"#const n={agents}.", f"agent(1..n). resource(1..x*y). utils({minval}..{maxval}).",
                f"#const x={x}. #const y={y}.",
                "edge(A,B) :- resource(A), resource(B), B == A + 1, A \ x != 0.",
                "edge(A,B) :- resource(A), resource(B), B == A + x."]
    for i in range(1, agents + 1):
        for j in range(1, x * y + 1):
            v = random.randrange(minval, maxval)
            encoding.append(f"util({i},{j},{v}).")
    encoding = '\n'.join(encoding)
    return encoding

def findMaxiMinNArySearch(instance, agent, denominator, length):
    start = time.time()
    models = []
    start = time.time()
    ctl = clingo.Control()
    ctl.load("../MaxiMin_Encodings/MaxMin_A_iterative_all.lp")
    ctl.add("base", [], instance)
    ctl.add("base", [], f"\nagentSelect({agent}).")
    ctl.ground([("base", [])])
    L = 0
    R = length
    while L < R:
        m = int(math.floor((R-L) / denominator + L))
        #print(f"M: {m}\n")
        ctl.ground([("value", [m])])
        ctl.assign_external(clingo.Function("val", [m]), True)
        models.append(modelResults())
        x = ctl.solve(on_model=models[-1].addResults)
        if x.unsatisfiable == True:
            R = m
        else:
            L = m + 1

        ctl.release_external(clingo.Function("val", [m]))
        ctl.cleanup()
    #print(f"T: {time.time() - start}\n")
    return (R - 1), (time.time() - start)

def findMaxiMinLinearSearch(instance, agent):
    m = []
    start = time.time()
    ctl = clingo.Control()
    ctl.load("../MaxiMin_Encodings/MaxMin_A_iterative_all.lp")
    ctl.add("base", [], instance)
    ctl.add("base", [], f"\nagentSelect({agent}).")
    ctl.ground([("base",[])])
    for i in range(1,100):
        ctl.ground([("value",[i])])
        ctl.assign_external(clingo.Function("val",[i]), True)
        m.append(modelResults())
        x = ctl.solve(on_model = m[i-1].addResults)
        if x.unsatisfiable == True:
            #print(f"T: {time.time() - start}\n")
            return i-1, (time.time() - start)
        ctl.release_external(clingo.Function("val",[i]))
        ctl.cleanup()


def findMaxiMin(instance, agent):
    start = time.time()
    ctl = clingo.Control()
    ctl.load("../MaxiMin_Encodings/MaxMin_A_mod.lp")
    ctl.add("base", [], instance)
    ctl.add("base", [], f"\nagentSelect({agent}).")
    ctl.ground([("base", [])])
    m = modelResults()
    ctl.solve(on_model=m.addResults)
    #print(f"T: {time.time() - start}\n")
    return -m.cost[0], (time.time() - start)





def solve(i, agents):
    d = {"Instance": i}
    values = []
    start_time = time.time()
    for agent in range(1, agents + 1):
        x = findMaxiMin(i, agent)
        values.append(-x.cost[-1])
        d[f"Agent{agent}_Cost"] = -x.cost[-1]
        # d[f"Agent{agent}_Symbols"] = x.symbols
    ctl = clingo.Control()
    ctl.load("/home/william/Fair_Division/T3/asp_encodings/maximin/MaxMin_B.lp")
    ctl.add("base", [], i)
    for agent in range(1, agents + 1):
        ctl.add("base", [], f"agentMin({agent},{values[agent - 1]}).")
    ctl.ground([("base", [])])
    m = modelResults()
    s = ctl.solve(on_model=m.addResults)
    d["Time"] = time.time() - start_time
    # d["Final"] = m
    d["SAT"] = s.satisfiable
    d["UNSAT"] = s.unsatisfiable
    return d


def generateAndSolveGrid(x, y, agents, minval, maxval):
    i = generateMaxiMinGrid(minval, maxval, agents, x, y)
    return solve(i, agents)


def generateAndSolveCycle(length, agents, minval, maxval):
    i = generateMaxiMinCycle(length, minval, maxval, agents)
    return solve(i, agents)


def generateAndSolveComplete(size, agents, minval, maxval):
    i = generateMaxiMinComplete(size, minval, maxval, agents)
    return solve(i, agents)
